''
'' FILE SEARCH - LOOKS FOR THE SPECIFIED FILE (MASKS SUPPORTED) RECURSIVELY,
'' STARTING FROM THE WORKING DIRECTORY. PRESS ANY KEY TO ABORT SEARCH.
'' MikeHawk, 2021 - Use at your own risk.
''

'$INCLUDE: 'QB.BI'
DEFINT A-Z

DECLARE SUB searchFile (where AS STRING, query AS STRING)
DECLARE SUB getObjects (where AS STRING, query AS STRING)
DECLARE FUNCTION getFileSize$ (l AS LONG, pad AS INTEGER)
DECLARE FUNCTION getTimeStamp$ (source AS STRING)
DECLARE FUNCTION getThisPath$ ()
DECLARE FUNCTION scopeDown% ()
DECLARE FUNCTION scopeUp$ ()
DECLARE FUNCTION scopeAppend% (folder AS STRING)
DECLARE FUNCTION INSTRREV% (source AS STRING, find AS STRING)

CONST moveROOT = 0
CONST moveDOWN = -1
CONST moveUP = 1

TYPE scopeInfo          ' OFS SZE   DESCRIPTION
	start AS INTEGER      ' 000 ..2   Offset of first folder in pathLst
	entry AS INTEGER      ' 002 ..2   Offset of last visited folder
END TYPE                ' 4 BYTES - Scope descriptor

DIM SHARED pathLst AS STRING * 20000      ' Parent and current sub-folders
DIM SHARED pathOfs AS INTEGER             ' Writing offset for pathLst
DIM SHARED levlNfo(0 TO 63) AS scopeInfo  ' Root plus 63 levels deep
DIM SHARED levlNow AS INTEGER             ' How deep we are
DIM SHARED resStr AS STRING               ' Result string
DIM SHARED resCnt AS INTEGER              ' Result count

DIM findMe AS STRING

findMe = COMMAND$
IF LEN(findMe) THEN
	searchFile ".", findMe
ELSE
	PRINT "USAGE: FindMe <filename>"
END IF

''
'' Add comma every three digits
''
FUNCTION getFileSize$ (l AS LONG, pad AS INTEGER)
	DIM src AS STRING, dst AS STRING, c AS INTEGER

	src = LTRIM$(STR$(l))
	DO
		IF (LEN(src) > 3) THEN c = 3 ELSE c = LEN(src)
		dst = RIGHT$(src, c) + dst
		src = LEFT$(src, LEN(src) - c)
		IF LEN(src) THEN dst = "," + dst
	LOOP WHILE LEN(src)
	getFileSize$ = RIGHT$(SPACE$(pad) + dst, pad)
END FUNCTION

''
'' Get all objects in current directory. This routine does it in two steps:
'' first it search for directories and then it searches for files according
'' to the specified <query>.
''
SUB getObjects (where AS STRING, query AS STRING)
	DIM DTA AS STRING * 44, MaskZ AS STRING
	DIM regs AS RegTypeX, objName AS STRING

	'' SETUP DTA SO WE DON'T DESTROY COMMAND$ ''
	regs.ax = &H1A00                  ' Set DTA function
	regs.dx = VARPTR(DTA)             ' DS:DX points to our DTA
	regs.ds = -1                      ' Use current value for DS
	CALL INTERRUPTX(&H21, regs, regs) ' Do the interrupt

	'' PARSE ALL OBJECTS, KEEP DIRECTORIES ''
	MaskZ = where + "\*.*" + CHR$(0)  '  Mask (search all)
	regs.ax = &H4E00                  '  FindFirst
	regs.cx = 16                      '  Get all object types
	regs.dx = SADD(MaskZ)             '  DS:DX points to ASCIIZ file mask
	regs.ds = -1                      '  Use current DS
	DO
		CALL INTERRUPTX(&H21, regs, regs)    ' Do the interrupt
		IF (regs.flags AND &H1) THEN EXIT DO ' No object left
		' Folder, append to scope
		IF (ASC(MID$(DTA, &H15 + 1, 1)) AND &H10) THEN
			objName = MID$(DTA, 31, INSTR(31, DTA, CHR$(0)) - 31)
			IF ((objName <> ".") AND (objName <> "..")) THEN
				IF scopeAppend%(objName) THEN PRINT : PRINT "Buffer overflow!": END
			END IF
		END IF
		regs.ax = &H4F00
	LOOP

	'' PARSE OBJECTS USING QUERY MASK ''
	MaskZ = where + "\" + query + CHR$(0)
	regs.ax = &H4E00
	regs.cx = 16
	regs.dx = SADD(MaskZ)
	regs.ds = -1
	DO
		CALL INTERRUPTX(&H21, regs, regs)
		IF (regs.flags AND &H1) THEN EXIT DO
		' File match
		IF ((ASC(MID$(DTA, &H15 + 1, 1)) AND &H10) = 0) THEN
			objName = where + "\" + MID$(DTA, 31, INSTR(31, DTA, CHR$(0)) - 31) + MID$(DTA, 23, 8)
			resStr = resStr + CHR$(LEN(objName)) + objName
			resCnt = resCnt + 1
		END IF
		regs.ax = &H4F00
	LOOP
END SUB

''
'' Get working directory
''
FUNCTION getThisPath$
	DIM path AS STRING * 65, regs AS RegTypeX, drive AS INTEGER
	DIM tempStr AS STRING

	regs.ax = &H1900        ' get active drive (likely "2" for "C:")
	CALL INTERRUPTX(&H21, regs, regs)
	drive = (regs.ax AND &HFF)

	regs.ax = &H4700        ' get current directory
	regs.dx = drive + 1     ' from active drive (0 to use default)
	regs.ds = -1            ' default segment
	regs.si = VARPTR(path)  ' string offset
	CALL INTERRUPTX(&H21, regs, regs)

	tempStr = LEFT$(path, INSTR(path, CHR$(0)) - 1)
	IF LEN(tempStr) THEN
		getThisPath$ = CHR$(drive + 65) + ":\" + tempStr
	ELSE
		getThisPath$ = CHR$(drive + 65) + ":"
	END IF
END FUNCTION

''
'' Convert 4-byte string to date & time
''
FUNCTION getTimeStamp$ (source AS STRING)
	DIM l AS LONG, tmp AS STRING

	l = CVL(RIGHT$(source, 2) + STRING$(2, 0))' date (YEAR/MONTH/DAY)
	tmp = LTRIM$(STR$(1980 + (l \ &H200))) + "-" + RIGHT$("00" + LTRIM$(STR$((l \ &H20) AND &HF)), 2) + "-" + RIGHT$("00" + LTRIM$(STR$(l AND &H1F)), 2)
	l = CVL(LEFT$(source, 2) + STRING$(2, 0))' time (HOURS:MIN)
	getTimeStamp$ = tmp + "  " + RIGHT$("00" + LTRIM$(STR$(l \ &H800)), 2) + ":" + RIGHT$("00" + LTRIM$(STR$((l \ &H20) AND &H3F)), 2)
END FUNCTION

''
'' Returns the position of the last occurence of <find> in <source>. INSTR
'' returns the position of the first time <find> is found inside <source>,
'' starting from <ofs + 1>. If <find> isn't spotted, INSTR will return 0. It
'' means that <ofs> will get bigger every time <find> is spotted, until <find>
'' is no longer found in <source> (at that point, <ofs> will be 0.) Since the
'' function is fed <ofs> before invoking INSTR, the function always contains
'' the last time <find> was found in <source>.
''
FUNCTION INSTRREV% (source AS STRING, find AS STRING)
	DIM ofs AS INTEGER
	DO
		INSTRREV% = ofs
		ofs = INSTR(ofs + 1, source, find)
	LOOP WHILE ofs
END FUNCTION

''
'' Append subfolder to scope. This function manipulates the string buffer that
'' contains every folder name <pathLst> and it's write cursor <pathOfs>. The
'' first byte provides the length in characters of the folder name, followed
'' by the folder name itself. This function also manipulates scope information
'' in array levlNfo(). <.start> is the offset of the first folder name in
'' <pathLst>, <.entry> is the offset of the last visited folder in that scope.
'' This function returns -1 if the memory buffer is saturated. Returns 0 if
'' the sub-folder was successfully added.
''
FUNCTION scopeAppend% (folder AS STRING)
	DIM tmp AS STRING

	IF (levlNfo(levlNow).start = 0) THEN
		levlNfo(levlNow).start = pathOfs
		levlNfo(levlNow).entry = pathOfs
	END IF

	tmp = LTRIM$(RTRIM$(UCASE$(folder)))
	tmp = CHR$(LEN(tmp)) + tmp
	IF ((pathOfs + LEN(tmp)) >= LEN(pathLst)) THEN
		scopeAppend% = -1
	ELSE
		MID$(pathLst, pathOfs, LEN(tmp)) = tmp
		pathOfs = pathOfs + LEN(tmp)
		scopeAppend% = 0
	END IF
END FUNCTION

''
'' Free current scope. Since folders are written by scope level (things to the
'' left of the string are parents, stuff to the right of the string are
'' children,) it means that we can just move the write cursor <pathOfs> back
'' to "erase" folders we already processed. This function also clears the
'' scope information entry so it can safely be re-used. This function returns
'' -1 if root as been depleted (nothing left to parse,) 0 otherwise.
''
FUNCTION scopeDown%
	IF levlNfo(levlNow).start THEN pathOfs = levlNfo(levlNow).start
	levlNfo(levlNow).start = 0
	levlNfo(levlNow).entry = 0
	scopeDown% = (levlNow = 0)
END FUNCTION

''
'' scopeUp$ returns the next sub-folder in this scope we should be browsing.
'' If no folder is available, scopeUp$ will return a blank string and you
'' should scopeDown% instead.
''
FUNCTION scopeUp$
	DIM offset AS INTEGER, length AS INTEGER

	' Nothing left to see here
	IF (levlNfo(levlNow).entry > pathOfs) OR (levlNfo(levlNow).entry = 0) THEN
		scopeUp$ = ""
	' Get next folder in line to browse
	ELSE
		offset = levlNfo(levlNow).entry
		length = ASC(MID$(pathLst, offset, 1))
		offset = offset + length + 1
		scopeUp$ = MID$(pathLst, offset - length, length)
		levlNfo(levlNow).entry = offset
	END IF
END FUNCTION

''
'' <where> is the base directory (where we start looking for the file,) it is
'' ideally an absolute path and shouldn't be terminated with a slash. <query>
'' is the mask of the object we're looking for.
''
SUB searchFile (where AS STRING, query AS STRING)
	DIM pathNxt AS STRING, pathAll AS STRING, move AS INTEGER
	DIM resOfs AS INTEGER, resLen AS INTEGER
	DIM spin AS INTEGER, pathLoc AS STRING

	pathOfs = 1     ' Always set to 1 before starting
	move = moveROOT ' Pretend we're moving up (for folder list)
	levlNow = 0     ' We start at level 0
	resCnt = 0      ' Reset result count
	resStr = ""     ' Reset result string

	pathAll = LTRIM$(RTRIM$(UCASE$(where)))
	IF (RIGHT$(pathAll, 1) = "\") THEN pathAll = LEFT$(pathAll, LEN(pathAll) - 1)
	query = LTRIM$(RTRIM$(UCASE$(query)))

	PRINT "Searching...  ";

	DO
		IF (move <> moveDOWN) THEN
			LOCATE , POS(0) - 1: PRINT MID$("-\|/", 1 + spin \ 4, 1);
			spin = ((spin + 1) AND &HF)

			' Append directories, search for file
			IF (move) THEN
				levlNow = levlNow + 1
				IF (levlNow > UBOUND(levlNfo)) THEN PRINT : PRINT "Scope overflow!": EXIT DO
			END IF
			CALL getObjects(pathAll, query)
		END IF

		' Enter next sub-directory
		pathNxt = scopeUp$
		IF LEN(pathNxt) THEN
			pathAll = pathAll + "\" + pathNxt
			move = moveUP
		ELSE
			IF (scopeDown%) THEN EXIT DO
			levlNow = levlNow - 1
			pathAll = LEFT$(pathAll, INSTRREV%(pathAll, "\") - 1)
			move = moveDOWN
		END IF
	LOOP UNTIL LEN(INKEY$)

	LOCATE , 1: PRINT LTRIM$(STR$(resCnt)); " file";
	IF (resCnt > 1) THEN PRINT "s";
	PRINT " found."
	IF LEN(resStr) THEN
		pathLoc = getThisPath$
		resOfs = 1
		DO
			resLen = ASC(MID$(resStr, resOfs, 1))
			PRINT TAB(3); RIGHT$(pathLoc + "\" + MID$(resStr, resOfs + 3, resLen - 10), 44); TAB(49); getFileSize$(CVL(MID$(resStr, resOfs + 1 + resLen - 4, 4)), 11); TAB(62); getTimeStamp$(MID$(resStr, resOfs + 1 + resLen - 8, 4))
			resOfs = resOfs + 1 + resLen
		LOOP WHILE (resOfs < LEN(resStr))
	END IF
END SUB

